#define _GNU_SOURCE 

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>

#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <curl/curl.h>

#include "strata_client.h"

struct reportfile {
  char* tmpdir;
  xmlTextWriterPtr writer;
  xmlBufferPtr buf;
  int issignature;
};

static char* internal_error_string = NULL;

static char*
ssprintf( const char* format, ... ) {
  size_t r;
  va_list p;
  char* retval;

  va_start(p, format);
  r = vasprintf(&retval, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("ssprintf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "ssprintf: vasprintf failed";
    return 0;
  }

  return retval;
}

static void
internal_error_printf( const char* format, ... ) {
  size_t r;
  va_list p;

  va_start(p, format);
  r = vasprintf(&internal_error_string, format, p);
  va_end(p);

  if (r < 0) {
    internal_error_string = strdup("internal_error_printf: vasprintf failed");
    if (!internal_error_string) 
      internal_error_string = "internal_error_printf: vasprintf failed";
  } 
}

static int 
run_command( const char *cmd )
{
    int retcode = system(cmd);
    if (retcode) {
      internal_error_printf("'%s' exited with status %d", cmd, retcode);
      return -1;
    }
}

static int
copy_file( const char *src_name, const char *dst_name )
{
  char* cmd = ssprintf("cp %s %s", src_name, dst_name);
  if (!cmd)
    return -1;
  return run_command(cmd);
}

static char*
concat_path_file(const char *path, const char *filename) {
  // This adds concats the two names, adding or removing one slash
  // as necessary.  It does not handle the case that path ends
  // in, or filename starts with, multiple slashs.

  const char* path_end = path + strlen(path) - 1;
  if ((*path_end == '/') != (*filename == '/'))
    // one has it, the other doesn't
    return ssprintf("%s%s", path, filename);
  else if (*filename == '/')
    // they both have it
    return ssprintf("%s%s", path, filename+1);
  else
    // neither has it
    return ssprintf("%s/%s", path, filename);
}



static int
write_str_to_file( const char *str, const char *path, const char *fname ) {

  char* ofile_name = concat_path_file(path, fname);
  if (!ofile_name)
    return -1;

  FILE *ofile = fopen(ofile_name, "w");
  if (!ofile) {
    internal_error_printf("Can't open '%s'", ofile_name);
    return -1;
  }

  fprintf(ofile, "%s\n", str);
  fclose(ofile);
  return 0;
}

const char* 
strata_client_strerror( void ) {
  return internal_error_string;
}

//
// The purpose of all the reportfile_XXX functions is to 
// collect up the contents of a report or signature into a 
// single tarball or xml file.
// 


// 
// This allocates a reportfile_t structure and initializes it.
//
//   Set 'issignature' to true if you only want an xml file.
//   If 'issignature' is true, then any bindings which would 
//   be written to an external file are dropped.
//
reportfile_t* 
reportfile_start( int issignature ) {

  // create a new reportfile_t
  reportfile_t* file = (reportfile_t*)malloc(sizeof(reportfile_t));

  file->issignature = issignature;

  // make a temp directory to work in
  file->tmpdir = strdup("/tmp/reportfileXXXXXX");
  if (mkdtemp(file->tmpdir) == NULL) {
    internal_error_printf("Can't mkdir a temporary directory in /tmp");
    return NULL;
  }

  // create a 'content' dir within temp directory
  char* content_dir_name = concat_path_file(file->tmpdir, "content");
  if (!content_dir_name)
    return NULL;
  if (mkdir(content_dir_name, 0700)) {
    internal_error_printf("Can't mkdir '%s'\n", content_dir_name);
    return NULL;
  }

  // set up a libxml 'buffer' and 'writer' to that buffer
  file->buf = xmlBufferCreate();
  if (file->buf == NULL) {
    internal_error_printf("strata_client: Error creating the xml buffer\n");
    return NULL;
  }
  file->writer = xmlNewTextWriterMemory(file->buf, 0);
  if (file->writer == NULL) {
    internal_error_printf("strata_client: Error creating the xml writer\n");
    return NULL;
  }

  // start a new xml document
  int rc;
  rc = xmlTextWriterStartDocument(file->writer, NULL, NULL, NULL);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartDocument\n");
    return NULL;
  }

  // start the document element 'report'
  rc = xmlTextWriterStartElement(file->writer, "report");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return NULL;
  }

  // add the 'xmlns' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, "xmlns",
                                   "http://www.redhat.com/gss/strata");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return NULL;
  }

  free(content_dir_name);
  return file;
}


static int
internal_reportfile_start_binding( reportfile_t* file, char* name, int isbinary, char* filename ) {
  int rc;

  // open a binding element
  rc = xmlTextWriterStartElement(file->writer, "binding");
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterStartElement\n");
    return -1;
  }

  // add the 'name' attribue
  rc = xmlTextWriterWriteAttribute(file->writer, "name", name);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterWriteAttribute\n");
    return -1;
  }


  // add the 'filename' attribute if neede
  if (filename) {
    rc = xmlTextWriterWriteAttribute(file->writer, "fileName", filename);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }

  // add the 'type' attribute
  if (isbinary) {
    rc = xmlTextWriterWriteAttribute(file->writer, "type", "binary");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, "type", "text");
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
}

//
// Add a new binding to a report, for a binding whos value is represented
//   as a string.
//
int
reportfile_add_binding_from_string( reportfile_t* file, const char* name, const char* value, int isbinary, const char* filename ) {
  int rc;

  // start the binding element with primary attributes
  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;


  // write the value of the binding, in the way that makes sense
  //   given it's value
  if (isbinary) {
    char* content_dir_name = concat_path_file(file->tmpdir, "content");
    if (!content_dir_name)
      return -1;
    rc = write_str_to_file(value, content_dir_name, name);
    if (rc < 0)
      return rc;
    free(content_dir_name);

    char* href_name = concat_path_file("content", name);
    if (!href_name)
      return -1;
    rc = xmlTextWriterWriteAttribute(file->writer, "href", content_dir_name);
    free(href_name);
    if (rc < 0) {
      internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }
  }
  else {
    rc = xmlTextWriterWriteAttribute(file->writer, "value", value);
      if (rc < 0) {
        internal_error_printf
	  ("strata_client: Error at xmlTextWriterWriteAttribute\n");
        return -1;
    }
  }

  rc = xmlTextWriterEndElement(file->writer);
  if (rc < 0) {
    internal_error_printf
      ("strata_client: Error at xmlTextWriterEndElement\n");
    return -1;
  }
}

//
// Add a new binding to a report, for a binding whos value is represented
//   as a named file.
//
//   'name' is the name of the binding
//   'local_filename' is the name of the file containing the bindings value
//   'filename' is the original name of the file, 
//      'filename' and 'local_filename' can be the same, but don't need to
//         be the same.
//
int
reportfile_add_binding_from_namedfile( reportfile_t* file, const char* name, const char* local_filename, int isbinary, const char* filename ) {
  int rc;

  rc = internal_reportfile_start_binding(file, name, isbinary, filename);
  if (rc < 0)
    return rc;

  char* content_dir_name = concat_path_file(file->tmpdir, "content");
  if (!content_dir_name)
    return -1;
  char* ofile_name = concat_path_file(content_dir_name, name);
  if (!ofile_name)
    return -1;
  rc = copy_file(local_filename, ofile_name);
  if (rc < 0)
    return rc;
  free(content_dir_name);
  free(ofile_name);

  
  char* href_name = concat_path_file("content", name);
  if (!href_name)
    return -1;
  rc = xmlTextWriterWriteAttribute(file->writer, "href", href_name);
  if (rc < 0) {
    internal_error_printf
	("strata_client: Error at xmlTextWriterWriteAttribute\n");
      return -1;
    }

  free(href_name);
  return 0;
}

//
// End the reportfile, and prepare it for delivery. 
// No more bindings can be added after this.
//
int
reportfile_end( reportfile_t* file ) {

  int rc;

  // close off the end of the xml file
  rc = xmlTextWriterEndDocument(file->writer);
  if (rc < 0) {
    internal_error_printf("strata_client: Error at xmlTextWriterEndDocument\n");
    return -1;
  }

  xmlFreeTextWriter(file->writer);
  file->writer = NULL;

  return 0;
}

//
// Write the reportfile to 'outfile_name'
//
// Currently this only works if the reportfile is a Report.
//
int
reportfile_write_to_file( reportfile_t* file, const char* outfile_name ) {

  if (0) {
    internal_error_printf("error: reportfile_write_to_file called before reportfile_end");
    return -1;
  }

  if (file->issignature) {
    internal_error_printf("error: reportfile_write_to_file called on signature");
    return -1;
  }



  int rc;

  // write the contents of the xmlfile to 'content.xml'
  rc = write_str_to_file(file->buf->content, file->tmpdir, "content.xml");
  if (rc < 0)
    return rc;

  // tar up the contents
  char* cmd = ssprintf("tar --file=%s -C %s --create --gzip content content.xml", outfile_name, file->tmpdir);
  if (!cmd) 
    return -1;
  rc = run_command(cmd);
  if (rc < 0)
    return -1;
  free(cmd);

  return 0;
}


//
// Return the contents of the reportfile as a string.
//
// Currently only works if 'file' is a signature.
//
char* 
reportfile_as_string( reportfile_t* file ) {

  if (0) {
    internal_error_printf("error: reportfile_as_string called before reportfile_end");
    return NULL;
  }

  if (!file->issignature) {
    internal_error_printf("error: reportfile_as_string called on non-signature");
    return NULL;
  }

  return strdup(file->buf->content);
}



//
// free the reportfile
//
int
reportfile_free( reportfile_t* file ) {
  
  if (file) {
    int rc;
    char* cmd;

    // delete the temp dir
    cmd = ssprintf("rm -rf %s", file->tmpdir);
    if (!cmd) 
      return -1;
    rc = run_command(cmd);
    if (rc < 0)
      return -1;
    free(cmd);

    xmlFreeTextWriter(file->writer);
    xmlBufferFree(file->buf);
    free(file->tmpdir);
    free(file);
  }

  return 0;
}





static size_t 
sendfile_read(void *buffer, size_t size, size_t nmemb, void *userp)
{
  return fread(buffer, size, nmemb, (FILE*)userp);
}

static void
internal_error_curl(CURLcode err, const char* msg1, const char* msg2)
{
  internal_error_printf("error: %s: %s: %s", msg1, msg2, curl_easy_strerror(err));
}


typedef struct response_data {
  long code;
  const char* location;
  const char* strata_message;
  char* body;
  size_t body_size;
} response_data_t;

// 
// Examine each header looking for "Location:" or "Strata-Message:" headers
// 
static size_t 
headerfunction(void *buffer_pv, size_t count, size_t nmemb, void *response_data_pv)
{
  size_t size = count * nmemb;
  struct response_data* response_data = (struct response_data*)response_data_pv;
  const char* buffer = (const char*)buffer_pv;

  const char location_key[] = "Location:";
  const size_t location_key_size = sizeof(location_key)-1;

  if (size >= location_key_size
      && 0 == memcmp(buffer,location_key,location_key_size)) {
    const char* start = (const char*) buffer+location_key_size+1;
    const char* end;

    // skip over any leading space
    while (start < buffer+size 
           && isspace(*start)) ++start;

    end = start;

    // skip till we find the end of the url (first space or end of buffer)
    while (end < buffer+size 
           && !isspace(*end)) ++end;

    response_data->location = strndup(start,end-start);
  }


  const char strata_message_key[] = "Strata-Message:";
  const size_t strata_message_key_size = sizeof(strata_message_key)-1;

  if (size >= strata_message_key_size
      && 0 == memcmp(buffer,strata_message_key,strata_message_key_size)) {
    const char* start = (const char*) buffer+strata_message_key_size+1;
    const char* end;

    // skip over any leading space
    while (start < buffer+size 
           && isspace(*start)) ++start;

    end = start;

    // skip till we find the end of the url (first space or end of buffer)
    while (end < buffer+size 
           && !isspace(*end)) ++end;

    response_data->strata_message = strndup(start,end-start);
  }

  return size;
}




static response_data_t* 
post(const char* URL,
     const char* data) {
  const char* retval;

  CURLcode rc;
  CURL *handle = curl_easy_init();
  //  FILE* file;
  struct curl_slist *httpheader_list = NULL;
  response_data_t* response_data = (response_data_t*)calloc(1,sizeof(response_data_t));

  rc = curl_easy_setopt(handle, CURLOPT_VERBOSE, 0);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_VERBOSE)");
    return NULL;
  }
  rc = curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_NOPROGRESS)");
    return NULL;
  }


  rc = curl_easy_setopt(handle, CURLOPT_POST, 1);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POST)");
    return NULL;
  }

  rc = curl_easy_setopt(handle, CURLOPT_URL, URL);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_URL)");
    return NULL;
  }
  
  httpheader_list = curl_slist_append(
      httpheader_list, 
      "Content-Type: application/xml");
  curl_easy_setopt(handle, CURLOPT_HTTPHEADER, httpheader_list);

  rc = curl_easy_setopt(handle, CURLOPT_POSTFIELDS, data);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_POSTFIELDS)");
    return NULL;
  }


  rc = curl_easy_setopt(handle, CURLOPT_HEADERFUNCTION, headerfunction);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_HEADERFUNCTION)");
    return NULL;
  }

  rc = curl_easy_setopt(handle, CURLOPT_WRITEHEADER, response_data );
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEHEADER)");
    return NULL;
  }

  FILE* body_stream = open_memstream(&response_data->body, &response_data->body_size); 
  rc = curl_easy_setopt(handle, CURLOPT_WRITEDATA, body_stream);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_setopt(CURLOPT_WRITEDATA)");
    return NULL;
  }

  rc = curl_easy_perform(handle);
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_perform");
    return NULL;
  }

  fclose(body_stream);

  rc = curl_easy_getinfo(handle, CURLINFO_RESPONSE_CODE, &response_data->code); 
  if (rc) {
    internal_error_curl(rc, "problem", "curl_easy_getinfo(CURLINFO_RESPONSE_CODE)");
    return NULL;
  }
 
  curl_easy_cleanup(handle);
  curl_slist_free_all(httpheader_list);

  return response_data;
}

const char* 
post_signature( const char* baseURL, const char* signature ) {
  const char* URL = concat_path_file(baseURL, "/signatures");
  const char* retval;

  response_data_t* response_data = post(URL, signature);
  if (!response_data)
    return NULL;

  switch (response_data->code) {
  case 200:
  case 201:
    if (response_data->body && strlen(response_data->body) > 0) {
      retval = response_data->body;
      response_data->body = NULL;
    }
    else if (response_data->strata_message 
             && strcmp(response_data->strata_message,"CREATED") != 0) {
      retval = response_data->strata_message;
      response_data->strata_message = NULL;
    }
    else
      retval = strdup("Thank you for submitting your signature.");
    break;
  default:
    if (response_data->strata_message) 
      retval = ssprintf("Error: %s (http code %ld)", 
                        response_data->strata_message, 
                        response_data->code );
    else
      retval = ssprintf("Error: Response Code: %ld\nBody:\n%s", response_data->code, response_data->body);
  }

  free((void*)response_data->strata_message);
  free((void*)response_data->body);
  free((void*)response_data->location);
  free((void *)response_data);
  free((void *)URL);
  return retval;
}

const char* 
create_case( const char* baseURL, const char* description ) {
  const char* URL = concat_path_file(baseURL, "/cases");
  const char* retval;

  response_data_t* response_data = post(URL, description);
  if (!response_data)
    return NULL;

  switch (response_data->code) {
  case 200:
  case 201:
    if (response_data->body && strlen(response_data->body) > 0) {
      retval = response_data->body;
      response_data->body = NULL;
    }
    else
      retval = ssprintf("Case Created: %s\n", response_data->location);
    break;
  default:
    if (response_data->strata_message) 
      retval = ssprintf("Error: %s (http code %ld)", 
                        response_data->strata_message, 
                        response_data->code );
    else
      retval = ssprintf("Error: Response Code: %ld\nBody:\n%s", response_data->code, response_data->body);
  }

  free((void*)response_data->strata_message);
  free((void*)response_data->body);
  free((void*)response_data->location);
  free((void *)response_data);
  free((void*)URL);
  return retval;
}

